'use strict';

var React = require('react');
var reactMotion = require('react-motion');
var PropTypes = require('prop-types');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = Object.assign || function __assign(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }
    return t;
};

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

var SizeAndPositionManager = function () {
    function SizeAndPositionManager(_a) {
        var itemCount = _a.itemCount,
            itemSizeGetter = _a.itemSizeGetter;
        this.itemSizeGetter = itemSizeGetter;
        this.itemCount = itemCount;
        // Cache of size and position data for items, mapped by item index.
        this.itemSizeAndPositionData = {};
        // Measurements for items up to this index can be trusted; items afterward should be estimated.
        this.lastMeasuredIndex = -1;
    }
    SizeAndPositionManager.prototype.updateConfig = function (_a) {
        var itemCount = _a.itemCount;
        this.itemCount = itemCount;
    };
    SizeAndPositionManager.prototype.getLastMeasuredIndex = function () {
        return this.lastMeasuredIndex;
    };
    /**
     * This method returns the size and position for the item at the specified index.
     * It just-in-time calculates (or used cached values) for items leading up to the index.
     */
    SizeAndPositionManager.prototype.getSizeAndPositionForIndex = function (index) {
        if (index < 0 || index >= this.itemCount) {
            throw Error("Requested index " + index + " is outside of range 0.." + this.itemCount);
        }
        if (index > this.lastMeasuredIndex) {
            var lastMeasuredSizeAndPosition = this.getSizeAndPositionOfLastMeasuredItem();
            var offset = lastMeasuredSizeAndPosition.offset + lastMeasuredSizeAndPosition.size;
            for (var i = this.lastMeasuredIndex + 1; i <= index; i++) {
                var size = this.itemSizeGetter(i);
                if (size == null || isNaN(size)) {
                    throw Error("Invalid size returned for index " + i + " of value " + size);
                }
                this.itemSizeAndPositionData[i] = {
                    offset: offset,
                    size: size
                };
                offset += size;
            }
            this.lastMeasuredIndex = index;
        }
        return this.itemSizeAndPositionData[index];
    };
    SizeAndPositionManager.prototype.getSizeAndPositionOfLastMeasuredItem = function () {
        return this.lastMeasuredIndex >= 0 ? this.itemSizeAndPositionData[this.lastMeasuredIndex] : { offset: 0, size: 0 };
    };
    /**
     * Total size of all items being measured.
     * This value will be completedly estimated initially.
     * As items as measured the estimate will be updated.
     */
    SizeAndPositionManager.prototype.getTotalSize = function () {
        if (this.itemCount > 0) {
            this.getSizeAndPositionForIndex(this.itemCount - 1);
        }
        var lastMeasuredSizeAndPosition = this.getSizeAndPositionOfLastMeasuredItem();
        return lastMeasuredSizeAndPosition.offset + lastMeasuredSizeAndPosition.size;
    };
    SizeAndPositionManager.prototype.getVisibleRange = function (_a) {
        var containerSize = _a.containerSize,
            offset = _a.offset,
            overscanCount = _a.overscanCount;
        var totalSize = this.getTotalSize();
        if (totalSize === 0) {
            return {};
        }
        var maxOffset = offset + containerSize;
        var start = this.findNearestItem(offset);
        if (typeof start === 'undefined') {
            throw Error("Invalid offset " + offset + " specified");
        }
        var datum = this.getSizeAndPositionForIndex(start);
        offset = datum.offset + datum.size;
        var stop = start;
        while (offset < maxOffset && stop < this.itemCount - 1) {
            stop++;
            offset += this.getSizeAndPositionForIndex(stop).size;
        }
        if (overscanCount) {
            start = Math.max(0, start - overscanCount);
            stop = Math.min(stop + overscanCount, this.itemCount - 1);
        }
        return {
            start: start,
            stop: stop
        };
    };
    /**
     * Clear all cached values for items after the specified index.
     * This method should be called for any item that has changed its size.
     * It will not immediately perform any calculations; they'll be performed the next time getSizeAndPositionForIndex() is called.
     */
    SizeAndPositionManager.prototype.resetItem = function (index) {
        this.lastMeasuredIndex = Math.min(this.lastMeasuredIndex, index - 1);
    };
    /**
     * Searches for the item (index) nearest the specified offset.
     *
     * If no exact match is found the next lowest item index will be returned.
     * This allows partially visible items (with offsets just before/above the fold) to be visible.
     */
    SizeAndPositionManager.prototype.findNearestItem = function (offset) {
        if (isNaN(offset)) {
            throw Error("Invalid offset " + offset + " specified");
        }
        // Our search algorithms find the nearest match at or below the specified offset.
        // So make sure the offset is at least 0 or no match will be found.
        offset = Math.max(0, offset);
        var lastMeasuredSizeAndPosition = this.getSizeAndPositionOfLastMeasuredItem();
        var lastMeasuredIndex = Math.max(0, this.lastMeasuredIndex);
        if (lastMeasuredSizeAndPosition.offset >= offset) {
            // If we've already measured items within this range just use a binary search as it's faster.
            return this.binarySearch({
                high: lastMeasuredIndex,
                low: 0,
                offset: offset
            });
        } else {
            // If we haven't yet measured this high, fallback to an exponential search with an inner binary search.
            // The exponential search avoids pre-computing sizes for the full set of items as a binary search would.
            // The overall complexity for this approach is O(log n).
            return this.exponentialSearch({
                index: lastMeasuredIndex,
                offset: offset
            });
        }
    };
    SizeAndPositionManager.prototype.binarySearch = function (_a) {
        var low = _a.low,
            high = _a.high,
            offset = _a.offset;
        var middle = 0;
        var currentOffset = 0;
        while (low <= high) {
            middle = low + Math.floor((high - low) / 2);
            currentOffset = this.getSizeAndPositionForIndex(middle).offset;
            if (currentOffset === offset) {
                return middle;
            } else if (currentOffset < offset) {
                low = middle + 1;
            } else if (currentOffset > offset) {
                high = middle - 1;
            }
        }
        if (low > 0) {
            return low - 1;
        }
        return 0;
    };
    SizeAndPositionManager.prototype.exponentialSearch = function (_a) {
        var index = _a.index,
            offset = _a.offset;
        var interval = 1;
        while (index < this.itemCount && this.getSizeAndPositionForIndex(index).offset < offset) {
            index += interval;
            interval *= 2;
        }
        return this.binarySearch({
            high: Math.min(index, this.itemCount - 1),
            low: Math.floor(index / 2),
            offset: offset
        });
    };
    return SizeAndPositionManager;
}();

var DIRECTION_VERTICAL = 'vertical';
var DIRECTION_HORIZONTAL = 'horizontal';
var scrollProp = (_a = {}, _a[DIRECTION_VERTICAL] = 'scrollTop', _a[DIRECTION_HORIZONTAL] = 'scrollLeft', _a);
var sizeProp = (_b = {}, _b[DIRECTION_VERTICAL] = 'height', _b[DIRECTION_HORIZONTAL] = 'width', _b);
var positionProp = (_c = {}, _c[DIRECTION_VERTICAL] = 'top', _c[DIRECTION_HORIZONTAL] = 'left', _c);
function noop() {} // tslint:disable-line no-empty
var _a;
var _b;
var _c;

var STYLE_WRAPPER = {
    overflow: 'auto',
    willChange: 'transform',
    WebkitOverflowScrolling: 'touch'
};
var STYLE_INNER = {
    position: 'relative',
    overflow: 'hidden',
    width: '100%',
    minHeight: '100%'
};
var VirtualList = function (_super) {
    __extends(VirtualList, _super);
    function VirtualList(props) {
        var _this = _super.call(this, props) || this;
        _this.sizeAndPositionManager = new SizeAndPositionManager({
            itemCount: _this.props.itemCount,
            itemSizeGetter: function (index) {
                return _this.getSize(index);
            }
        });
        _this.state = {
            offset: 0
        };
        _this.invisible = true;
        _this.lastOffsets = [];
        _this.lastRender = [];
        _this.getRef = function (node) {
            _this.rootNode = node;
            _this.props.getRef(node); // tslint:disable-line no-non-null-assertion
        };
        // this.willEnter = this.willEnter.bind(this);
        _this.willLeave = _this.willLeave.bind(_this);
        _this.willLeaveInvisible = _this.willLeaveInvisible.bind(_this);
        _this.handleScroll = _this.handleScroll.bind(_this);
        return _this;
    }
    VirtualList.prototype.componentWillReceiveProps = function (nextProps) {
        var _a = this.props,
            itemCount = _a.itemCount,
            itemSize = _a.itemSize;
        var itemPropsHaveChanged = nextProps.itemCount !== itemCount || nextProps.itemSize !== itemSize;
        if (nextProps.itemCount !== itemCount) {
            this.sizeAndPositionManager.updateConfig({
                itemCount: nextProps.itemCount
            });
        }
        if (itemPropsHaveChanged) {
            this.recomputeSizes();
        }
    };
    VirtualList.prototype.handleScroll = function (e) {
        var offset = this.getNodeOffset();
        if (offset < 0 || this.state.offset === offset || e.target !== this.rootNode) {
            return;
        }
        this.invisible = true;
        this.setState({
            offset: offset
        });
    };
    VirtualList.prototype.getNodeOffset = function () {
        var _a = this.props.scrollDirection,
            scrollDirection = _a === void 0 ? DIRECTION_VERTICAL : _a;
        return this.rootNode[scrollProp[scrollDirection]];
    };
    VirtualList.prototype.getSize = function (index) {
        var _a = this.props,
            itemSize = _a.itemSize,
            keyForIndex = _a.keyForIndex;
        if (typeof itemSize === 'function') {
            return itemSize(keyForIndex(index));
        }
        return Array.isArray(itemSize) ? itemSize[index] : itemSize;
    };
    VirtualList.prototype.willLeave = function (style) {
        var _a = this.props.scrollDirection,
            scrollDirection = _a === void 0 ? DIRECTION_VERTICAL : _a;
        var offset = this.lastOffsets[style.data];
        var offsetProp = positionProp[scrollDirection];
        style[offsetProp] = offset;
        return this.props.willLeave(style);
    };
    VirtualList.prototype.willLeaveInvisible = function (style) {
        var _a = this.props.scrollDirection,
            scrollDirection = _a === void 0 ? DIRECTION_VERTICAL : _a;
        var offset = this.lastOffsets[style.data];
        var offsetProp = positionProp[scrollDirection];
        style[offsetProp] = offset;
        return this.props.willLeaveInvisible(style);
    };
    VirtualList.prototype.recomputeSizes = function (startIndex) {
        if (startIndex === void 0) {
            startIndex = 0;
        }
        this.sizeAndPositionManager.resetItem(startIndex);
    };
    VirtualList.prototype.render = function () {
        var _this = this;
        var _a = this.props,
            height = _a.height,
            _b = _a.overscanCount,
            overscanCount = _b === void 0 ? 3 : _b,
            renderItem = _a.renderItem,
            itemCount = _a.itemCount,
            itemSize = _a.itemSize,
            _c = _a.scrollDirection,
            scrollDirection = _c === void 0 ? DIRECTION_VERTICAL : _c,
            style = _a.style,
            getRef = _a.getRef,
            getContainerRef = _a.getContainerRef,
            width = _a.width,
            defaultStyle = _a.defaultStyle,
            defaultStyleInvisible = _a.defaultStyleInvisible,
            willLeave = _a.willLeave,
            willLeaveInvisible = _a.willLeaveInvisible,
            willEnter = _a.willEnter,

        // didLeave,
        keyForIndex = _a.keyForIndex,
            props = __rest(_a, ["height", "overscanCount", "renderItem", "itemCount", "itemSize", "scrollDirection", "style", "getRef", "getContainerRef", "width", "defaultStyle", "defaultStyleInvisible", "willLeave", "willLeaveInvisible", "willEnter", "keyForIndex"]);
        var offset = this.state.offset;
        var _d = this,
            manager = _d.sizeAndPositionManager,
            invisible = _d.invisible;
        var _e = manager.getVisibleRange({
            containerSize: this.props[sizeProp[scrollDirection]] || 0,
            offset: offset,
            overscanCount: overscanCount
        }),
            start = _e.start,
            stop = _e.stop;
        var items = [];
        var loadedKeys = {};
        if (typeof start !== 'undefined' && typeof stop !== 'undefined') {
            var offsetProp = positionProp[scrollDirection];
            for (var index = start; index <= stop; index++) {
                var _f = manager.getSizeAndPositionForIndex(index),
                    size = _f.size,
                    offset_1 = _f.offset;
                this.lastOffsets[index] = offset_1;
                var key = keyForIndex(index);
                var style_1 = __assign({}, invisible ? defaultStyleInvisible(key, size) : defaultStyle(key, size), (_g = {}, _g[offsetProp] = offset_1, _g));
                loadedKeys[key] = true;
                items.push({
                    key: key,
                    data: index,
                    style: style_1
                });
            }
        }
        return React.createElement("div", __assign({ ref: this.getRef }, props, { onScroll: this.handleScroll, style: __assign({}, STYLE_WRAPPER, style, { height: height, width: width }) }), React.createElement(reactMotion.TransitionMotion, { willLeave: invisible ? this.willLeaveInvisible : this.willLeave, willEnter: willEnter,
            // didLeave={didLeave}
            styles: items }, function (styles) {
            return React.createElement("div", { ref: getContainerRef, style: __assign({}, STYLE_INNER, (_a = {}, _a[sizeProp[scrollDirection]] = _this.sizeAndPositionManager.getTotalSize(), _a)) }, styles.map(function (style) {
                var key = style.key;
                if (!loadedKeys[key]) {
                    return _this.lastRender[key];
                }
                return _this.lastRender[key] = renderItem(style);
            }));
            var _a;
        }));
        var _g;
    };
    VirtualList.prototype.componentDidMount = function () {
        this.invisible = false;
    };
    VirtualList.prototype.componentDidUpdate = function () {
        this.invisible = false;
    };
    VirtualList.defaultProps = {
        overscanCount: 3,
        scrollDirection: DIRECTION_VERTICAL,
        getContainerRef: noop,
        getRef: noop,
        width: '100%'
        // didLeave: () => {} // tslint:disable-line no-empty
    };
    VirtualList.propTypes = {
        height: PropTypes.oneOfType([PropTypes.number, PropTypes.string]).isRequired,
        itemCount: PropTypes.number.isRequired,
        itemSize: PropTypes.oneOfType([PropTypes.number, PropTypes.array, PropTypes.func]).isRequired,
        overscanCount: PropTypes.number,
        renderItem: PropTypes.func.isRequired,
        scrollDirection: PropTypes.oneOf([DIRECTION_HORIZONTAL, DIRECTION_VERTICAL]).isRequired,
        width: PropTypes.oneOfType([PropTypes.number, PropTypes.string]).isRequired
    };
    return VirtualList;
}(React.PureComponent);

module.exports = VirtualList;
